'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var api = require('@thi.ng/api');
var checks = require('@thi.ng/checks');
var errors = require('@thi.ng/errors');

const toPath = (path) => checks.isArray(path)
    ? path
    : checks.isString(path)
        ? path.length > 0
            ? path.split(".")
            : []
        : path != null
            ? [path]
            : [];
const exists = (obj, path) => {
    if (obj == null) {
        return false;
    }
    path = toPath(path);
    for (let n = path.length - 1, i = 0; i <= n; i++) {
        const k = path[i];
        if (!obj.hasOwnProperty(k)) {
            return false;
        }
        obj = obj[k];
        if (obj == null && i < n) {
            return false;
        }
    }
    return true;
};
const disallowProtoPath = (path) => (api.assert(!checks.isProtoPath(path), `unsafe path: '${path}'`), path);

const defGetterUnsafe = (path) => defGetter(path);
function defGetter(path) {
    const ks = toPath(path);
    const [a, b, c, d] = ks;
    switch (ks.length) {
        case 0:
            return (s) => s;
        case 1:
            return (s) => (s != null ? s[a] : undefined);
        case 2:
            return (s) => s != null ? ((s = s[a]) != null ? s[b] : undefined) : undefined;
        case 3:
            return (s) => s != null
                ? (s = s[a]) != null
                    ? (s = s[b]) != null
                        ? s[c]
                        : undefined
                    : undefined
                : undefined;
        case 4:
            return (s) => s != null
                ? (s = s[a]) != null
                    ? (s = s[b]) != null
                        ? (s = s[c]) != null
                            ? s[d]
                            : undefined
                        : undefined
                    : undefined
                : undefined;
        default:
            return (s) => {
                const n = ks.length - 1;
                let res = s;
                for (let i = 0; res != null && i <= n; i++) {
                    res = res[ks[i]];
                }
                return res;
            };
    }
}

const defSetterUnsafe = (path) => defSetter(path);
function defSetter(path) {
    const ks = toPath(path);
    const [a, b, c, d] = ks;
    switch (ks.length) {
        case 0:
            return (_, v) => v;
        case 1:
            return (s, v) => ((s = copy(s)), (s[a] = v), s);
        case 2:
            return (s, v) => {
                let x;
                s = copy(s);
                s[a] = x = copy(s[a]);
                x[b] = v;
                return s;
            };
        case 3:
            return (s, v) => {
                let x, y;
                s = copy(s);
                s[a] = x = copy(s[a]);
                x[b] = y = copy(x[b]);
                y[c] = v;
                return s;
            };
        case 4:
            return (s, v) => {
                let x, y, z;
                s = copy(s);
                s[a] = x = copy(s[a]);
                x[b] = y = copy(x[b]);
                y[c] = z = copy(y[c]);
                z[d] = v;
                return s;
            };
        default:
            let f;
            for (let i = ks.length; --i >= 0;) {
                f = compS(ks[i], f);
            }
            return f;
    }
}
const copy = (x) => checks.isArray(x) || checks.isTypedArray(x) ? x.slice() : Object.assign({}, x);
const compS = (k, f) => (s, v) => ((s = copy(s)), (s[k] = f ? f(s[k], v) : v), s);

const updateInUnsafe = (state, path, fn, ...args) =>
updateIn(state, path, fn, ...args);
function updateIn(state, path, fn, ...args) {
    return defSetter(path)(state,
    fn.apply(null, (args.unshift(defGetter(path)(state)), args)));
}

const deleteInUnsafe = (state, path) => deleteIn(state, path);
function deleteIn(state, path) {
    const ks = toPath(path).slice();
    if (ks.length) {
        const k = ks.pop();
        return updateIn(state, ks, (x) => ((x = Object.assign({}, x)), delete x[k], x));
    }
}

const getInUnsafe = (state, path) => defGetter(path)(state);
function getIn(state, path) {
    return defGetter(path)(state);
}

const defMutatorUnsafe = (path) => defMutator(path);
function defMutator(path) {
    const ks = toPath(path);
    disallowProtoPath(ks);
    let [a, b, c, d] = ks;
    switch (ks.length) {
        case 0:
            return (_, x) => x;
        case 1:
            return (s, x) => (s ? ((s[a] = x), s) : undefined);
        case 2:
            return (s, x) => {
                let t;
                return s
                    ? (t = s[a])
                        ? ((t[b] = x), s)
                        : undefined
                    : undefined;
            };
        case 3:
            return (s, x) => {
                let t;
                return s
                    ? (t = s[a])
                        ? (t = t[b])
                            ? ((t[c] = x), s)
                            : undefined
                        : undefined
                    : undefined;
            };
        case 4:
            return (s, x) => {
                let t;
                return s
                    ? (t = s[a])
                        ? (t = t[b])
                            ? (t = t[c])
                                ? ((t[d] = x), s)
                                : undefined
                            : undefined
                        : undefined
                    : undefined;
            };
        default:
            return (s, x) => {
                let t = s;
                const n = ks.length - 1;
                for (let k = 0; k < n; k++) {
                    if (!(t = t[ks[k]]))
                        return;
                }
                t[ks[n]] = x;
                return s;
            };
    }
}

const mutInUnsafe = (state, path, val) => defMutator(path)(state, val);
function mutIn(state, path, val) {
    return defMutator(path)(state, val);
}

function mutInManyUnsafe(state, ...pairs) {
    const n = pairs.length;
    n & 1 && errors.illegalArgs(`require even number of args (got ${pairs.length})`);
    for (let i = 0; i < n && state; i += 2) {
        state = mutIn(state, pairs[i], pairs[i + 1]);
    }
    return state;
}

const setInUnsafe = (state, path, val) => defSetter(path)(state, val);
function setIn(state, path, val) {
    return defSetter(path)(state, val);
}

function setInManyUnsafe(state, ...pairs) {
    const n = pairs.length;
    n & 1 &&
        errors.illegalArgs(`require even number of KV args (got ${pairs.length})`);
    for (let i = 0; i < n; i += 2) {
        state = setIn(state, pairs[i], pairs[i + 1]);
    }
    return state;
}

const defUpdaterUnsafe = (path, fn) => defUpdater(path, fn);
function defUpdater(path, fn) {
    const g = defGetter(path);
    const s = defSetter(path);
    return (state, ...args) => s(state, fn.apply(null, (args.unshift(g(state)), args)));
}

exports.copy = copy;
exports.defGetter = defGetter;
exports.defGetterUnsafe = defGetterUnsafe;
exports.defMutator = defMutator;
exports.defMutatorUnsafe = defMutatorUnsafe;
exports.defSetter = defSetter;
exports.defSetterUnsafe = defSetterUnsafe;
exports.defUpdater = defUpdater;
exports.defUpdaterUnsafe = defUpdaterUnsafe;
exports.deleteIn = deleteIn;
exports.deleteInUnsafe = deleteInUnsafe;
exports.disallowProtoPath = disallowProtoPath;
exports.exists = exists;
exports.getIn = getIn;
exports.getInUnsafe = getInUnsafe;
exports.mutIn = mutIn;
exports.mutInManyUnsafe = mutInManyUnsafe;
exports.mutInUnsafe = mutInUnsafe;
exports.setIn = setIn;
exports.setInManyUnsafe = setInManyUnsafe;
exports.setInUnsafe = setInUnsafe;
exports.toPath = toPath;
exports.updateIn = updateIn;
exports.updateInUnsafe = updateInUnsafe;
